package com.Felipe.Wave;

import java.io.ByteArrayInputStream;
import java.io.IOException;

import com.musicg.math.rank.ArrayRankDouble;
import com.musicg.math.statistics.StandardDeviation;
import com.musicg.math.statistics.ZeroCrossingRate;
import com.musicg.wave.Wave;
import com.musicg.wave.WaveHeader;
import com.musicg.wave.extension.Spectrogram;

/*
 * Vypis fingerprint z wave soubor [ureno pro krtk rzn zvuky]
 * 
 * Zdroje:
 * Kombinace nassledujcch 
 * "com.musicg.wave.WaveTypeDetector.getWhistleProbability()" a 
 * "com.musicg.api.DetectionApi"
 * 
 *  vechny metody jsou statick a pipraveny k pouit, bez poteby vytvet tdy instance
 */

public class DetektorVlastnosti {
	
	/*
	 * Vpis zavis na nsledujcch 4ech parametrech
	 * wave : zvukova vlna v "Wave" formatu
	 * lowpass a highpass
	 * robustnost k vpotu smrodatn odchylky
	 */
	public static ActionZvukovyOtisk extractFingerprint(Wave wave, int lowPass, int highPass, int robustness){
		ActionZvukovyOtisk fp = new ActionZvukovyOtisk();
		
		WaveHeader wavHeader = wave.getWaveHeader();
		
		//fft poadavky
		int fftSampleSize = 1024;
		int fftSignalByteLength = fftSampleSize * wavHeader.getBitsPerSample() / 8;
		
		byte[] audioBytes = wave.getBytes();
		ByteArrayInputStream inputStream = new ByteArrayInputStream(audioBytes);

		int numFrames = inputStream.available() / fftSignalByteLength; //divide wave in frames [these will be used to create finger print slices]
		byte[] bytes = new byte[fftSignalByteLength];
		
		//opakovn a pidn ezu otisku zvuku
		for (int frameNumber = 0; frameNumber < numFrames; frameNumber++) { //F
			try {
				//ten
				inputStream.read(bytes);
				
				fp.addSlice( extractFingerprintSlice(wavHeader, frameNumber, bytes, lowPass, highPass, robustness)); //add the slice
				
				
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		return fp;
	}
	public static RezOtiskuZvuku extractFingerprintSlice(WaveHeader waveHeader, int frameNumber, byte []audioBytes, int lowPass, int highPass, int robustness){
		RezOtiskuZvuku slice = new  RezOtiskuZvuku();
		
		slice.frameNumber = frameNumber;
		
		int bytesPerSample = waveHeader.getBitsPerSample() / 8;
		int numSamples = audioBytes.length / bytesPerSample;

		// numSamples umocnme na 2
		if (numSamples > 0 && Integer.bitCount(numSamples) == 1) {
			int fftSampleSize = numSamples;
			int numFrequencyUnit = fftSampleSize / 2;

			// frekvence mus bt zaznamenna v polovin z nSamples co je definovno v Nyquist theory
			double unitFrequency = (double) waveHeader.getSampleRate() / 2 / numFrequencyUnit;

			// nastaveni hranic
			int dolniHranice = (int) (highPass / unitFrequency);
			int horniHranice = (int) (lowPass / unitFrequency);
			
			// konec nastaveni hranic

			Wave wave = new Wave(waveHeader, audioBytes);	// bajty zvukovho snmku (framu)
			short[] amplitudes = wave.getSampleAmplitudes();

			// spektrum
			Spectrogram spectrogram = wave.getSpectrogram(fftSampleSize, 0);
			
			double[][] spectrogramData = spectrogram.getAbsoluteSpectrogramData();

			// dokud fftSampleSize==numSamples, pouv pouze jdno spektrum thisFrameSpectrogramData[0]
			double[] spectrum = spectrogramData[0];

			int frequencyUnitRange = horniHranice - dolniHranice + 1;		
			double[] rangedSpectrum = new double[frequencyUnitRange];
			System.arraycopy(spectrum, dolniHranice, rangedSpectrum, 0, rangedSpectrum.length);
			
			if (frequencyUnitRange <= spectrum.length) {
				//export parametr
				
				slice.intenzita = getIntensity(spectrum);
				slice.smerOdchylka = getStandardDeviation(spectrogramData, robustness);
				slice.ZeroCrossingRate = getZeroCrossingRate(amplitudes);
				slice.frekvence = getFrequency(rangedSpectrum, unitFrequency);
				
			} else {
				System.err
						.println("error: vlna musi mt vy sample rate");
			}

		} else {
			System.out.println("Velikost vzorku mus byt umocnna na druhou");
		}
		return slice;
	}
	protected static double getIntensity(double[] spectrum){
		// get prmr intenzity signlu
		double intensity = 0;
		for (int i = 0; i < spectrum.length; i++) {
			intensity += spectrum[i];
		}
		intensity /= spectrum.length;
		// konec get prmr intenzity signlu
		return intensity;
	}
	protected static void normalizeSpectrogramData(double[][] spectrogramData) {

		
		// set max a min aplitudy
		double maxAmp = Double.MIN_VALUE;
		double minAmp = Double.MAX_VALUE;
		for (int i = 0; i < spectrogramData.length; i++) {
			for (int j = 0; j < spectrogramData[i].length; j++){
				if (spectrogramData[i][j] > maxAmp) {
					maxAmp = spectrogramData[i][j];
				} else if (spectrogramData[i][j] < minAmp) {
					minAmp = spectrogramData[i][j];
				}
			}
		}
		// konec set max a min amplitudy

		
		// vyhnut se dlen nulou
		double minValidAmp = 0.00000000001F;
		if (minAmp == 0) {
			minAmp = minValidAmp;
		}

		double diff = Math.log10(maxAmp / minAmp); 
		for (int i = 0; i < spectrogramData.length; i++) {
			for (int j = 0; j < spectrogramData[i].length; j++) {
				if (spectrogramData[i][j] < minValidAmp) {
					spectrogramData[i][j] = 0;
				} else {
					spectrogramData[i][j] = (Math.log10(spectrogramData[i][j] / minAmp)) / diff;
				}
			}
		}
		
	}
	
protected static double getStandardDeviation(double[][] spectrogramData, int robustness){
		
		
		normalizeSpectrogramData(spectrogramData);
			
		// analyza dat ve framu
		// dokud fftSampleSize==numSamples, pouva se pouze jeden spektogram spectrogramData[last]
		double[] spectrum = spectrogramData[spectrogramData.length - 1];
		// hledn nejrobutnjch frekvenc ve framu
		double[] robustFrequencies = new double[robustness];
		ArrayRankDouble arrayRankDouble = new ArrayRankDouble();
		double nthValue = arrayRankDouble.getNthOrderedValue(spectrum, robustness, false);
		// end analyzy dat ve framu
				
		int count = 0;
		for (int i = 0; i < spectrum.length; i++) {
			if (spectrum[i] >= nthValue) {
				robustFrequencies[count++] = spectrum[i];
				if (count >= robustness) {
					break;
				}
			}
		}
		// end hledani nejrobustnejsich frekvenci

		StandardDeviation standardDeviation = new StandardDeviation();
		standardDeviation.setValues(robustFrequencies);
		double sd = standardDeviation.evaluate();
		return sd;
	}
protected static int getZeroCrossingRate(short[] amplitudes){
	ZeroCrossingRate zcr = new ZeroCrossingRate(amplitudes, 1);
	int numZeroCrosses = (int) zcr.evaluate();
	
	// rozdilne zvuky maji rozdilnou hodnotu pocetu pruchudu nulou zerocrossing rate
	// pokud lengthInSecond=1, zero crossing rate se rovna  poet pechodu pes nulu
	
	return numZeroCrosses;
}
protected static double getFrequency(double[] spectrum, double unitFrequency){
	// hledani robustni frekvence
	ArrayRankDouble arrayRankDouble = new ArrayRankDouble();
	double robustFrequency = arrayRankDouble.getMaxValueIndex(spectrum) * unitFrequency;
	return robustFrequency;
}
}

